from lotgi.models import rest

import logging
import requests
from typing import List, Optional
import os
import tomli
import functools

logger = logging.getLogger(__name__)

@functools.cache
def get_cloud_credentials():
    modal_credentials = None
    try:
        modal = tomli.load(open(f"{os.environ.get('HOME', '')}/.modal.toml", "rb"))
        modal_credentials = rest.ModalCredentials(
            **modal["default"]
        )
    except Exception as e:
        logger.warning(f"Couldn't read modal credentials. {e}")

    runpod_credentials = None
    try:
        runpod_credentials = rest.RunpodCredentials(
            api_key=os.environ["RUNPOD_API_KEY"]
        )
    except Exception as e:
        logger.warning(f"Couldn't read runpod credentials. {e}")

    return rest.CloudCredentials(
        modal_credentials=modal_credentials,
        runpod_credentials=runpod_credentials
    )

def get_huggingface_credentials():
    try:
        home_dir = os.environ["HOME"]
        with open(f"{home_dir}/.cache/huggingface/token", "r") as f:
            token = f.read()
            return token.strip()
    except Exception as e:
        logger.warning(f"Didn't detect huggingface credentials. Only public models with work. Run `huggingface-cli login` to set your huggingface credentials. {e}")

class RestClient:
    def __init__(self):
        """
        """
        self.endpoint = os.environ.get("LOTGI_ENDPOINT", "https://lotgi.ai/api")
        # TODO: User should specify some secret token not user id, duh.
        if "LOTGI_TOKEN" not in os.environ:
            raise KeyError(f"LOTGI_TOKEN not set.")

        self._token = os.environ["LOTGI_TOKEN"]

        self.headers = {
            "lotgi-token": self._token,
            "Content-type": "application/json",
        }

    def submit_job(self, *,
                   model : str,
                   input_file : str,
                   field : str,
                   target_cost : float,
                   target_deadline : int,
                   execution_tradeoff : str,
                   max_output_tokens : Optional[int],
                   regex : Optional[str],
                   _autodetect_cloud_credentials : bool = False
                   ) -> rest.JobStatus:
        url = f"{self.endpoint}/create_job"

        if _autodetect_cloud_credentials:
            cloud_credentials = get_cloud_credentials()
        else:
            cloud_credentials = None


        if max_output_tokens is None:
            max_output_tokens = rest.DEFAULT_MAX_OUTPUT_TOKENS
        if regex  is None:
            regex = rest.DEFAULT_REGEX

        job_spec = rest.JobSpec(
            cloud_credentials=cloud_credentials,
            input_file=input_file,
            model=model,
            field=field,
            target_cost=target_cost,
            target_deadline=target_deadline,
            execution_tradeoff=execution_tradeoff,
            huggingface_token=get_huggingface_credentials(),
            max_output_tokens=max_output_tokens,
            regex=regex,
        )


        body = job_spec.dict()

        result = requests.post(url, headers=self.headers, json=body)

        if result.status_code == 401:
            raise PermissionError(f"Request authentication failed. Is your `LOTGI_TOKEN` environment variable set correctly? {result.status_code} : {result.json()['detail']}")

        if not result.ok:
            raise RuntimeError(f"Request was not successfully submitted. {result.status_code}: {result.text}")

        return rest.JobStatus.validate(result.json())

    def job_exection_options(self, *, job_execution_options_input : rest.JobExecutionOptionsInput) -> List[rest.JobExecutionTradeoffOption]:
        url = f"{self.endpoint}/job_execution_options"

        result = requests.get(url, headers=self.headers, json=job_execution_options_input.dict())

        if not result.ok:
            raise RuntimeError(f"Request was not successfully submitted. {result.status_code}: {result.text}")

        return [
            rest.JobExecutionTradeoffOption.parse_obj(result.json()[i])
            for i in range(len(result.json()))
        ]


    def list_jobs(self) -> List[rest.JobStatus]:
        url = f"{self.endpoint}/list_jobs"
        result = requests.get(url, headers=self.headers)

        if result.status_code == 401:
            raise PermissionError(f"Request authentication failed. Is your `LOTGI_TOKEN` environment variable set correctly? {result.status_code} : {result.json()['detail']}")

        if not result.ok:
            raise RuntimeError(f"Request was not successfully submitted. {result.status_code}: {result.text}")

        return [
            rest.JobStatus.validate(status)
            for status in result.json()
        ]

    def get_job_results(self, job_id : str) -> str:
        url = f"{self.endpoint}/get_job_results"
        result = requests.get(url, headers=self.headers, params={"job_id": job_id})

        if result.status_code == 401:
            raise PermissionError(f"Request authentication failed. Is your `LOTGI_TOKEN` environment variable set correctly? {result.status_code} : {result.json()['detail']}")

        if result.status_code == 404:
            raise KeyError(f"Job doesn't exist! {result.text}")

        if not result.ok:
            raise RuntimeError(f"Request was not successfully submitted. {result.status_code}: {result.text}")
        return result.json()

    def get_upload_url(self, filename : str) -> str:
        url = f"{self.endpoint}/get_upload_url"
        result = requests.get(url, headers=self.headers, params={"filename": filename})

        if result.status_code == 401:
            raise PermissionError(f"Request authentication failed. Is your `LOTGI_TOKEN` environment variable set correctly? {result.status_code} : {result.json()['detail']}")

        if not result.ok:
            raise RuntimeError(f"Request was not successfully submitted. {result.status_code}: {result.text}")
        return result.json()



from pydantic import BaseModel, Field, model_validator, Extra
from typing import Optional
import shortuuid

uuid_generator = shortuuid.ShortUUID(alphabet="abcdefghijklmnopqrstuvwxyz1234567890")

class ModalCredentials(BaseModel):
    token_id : str
    token_secret : str

class RunpodCredentials(BaseModel):
    api_key : str

class CloudCredentials(BaseModel):
    modal_credentials : Optional[ModalCredentials] = None
    runpod_credentials : Optional[RunpodCredentials] = None

    @model_validator(mode="after")
    def atleast_one_credential(self):
        if not (self.modal_credentials or self.runpod_credentials):
            raise ValueError("At least one set of credentials should be specified.")
        return self

EXECUTION_TRADEOFFS = {
    "Cheapest",
    "Hour",
}

DEFAULT_MAX_OUTPUT_TOKENS = 32
DEFAULT_REGEX = r".*"

class JobSpec(BaseModel):
    job_id : str = Field(default_factory=lambda : f"job-{uuid_generator.uuid()}")
    cloud_credentials : Optional[CloudCredentials] = None
    input_file : str
    target_cost : float
    target_deadline : int # Specified as a unix timestamp
    model : str
    field : str
    execution_tradeoff : str
    max_output_tokens : int = DEFAULT_MAX_OUTPUT_TOKENS
    regex : str = DEFAULT_REGEX
    huggingface_token : Optional[str] = None

    # @model_validator(mode="after")
    # def valid_execution_tradeoff(self):
    #     if self.execution_tradeoff not in EXECUTION_TRADEOFFS:
    #         raise ValueError(f"execution_tradeoff must be one off {EXECUTION_TRADEOFFS}")
    #     return self

class JobStatus(BaseModel, extra=Extra.allow):
    job_id : str
    state : str
    model : str
    submission_time : int
    execution_tradeoff : str
    input_file : str
    target_cost : float
    target_deadline : int # Specified as a unix timestamp
    field : str

class JobExecutionOptionsInput(BaseModel):
    input_token_count : int
    num_requests : int
    model : str
    max_output_tokens : int = 32

class JobExecutionTradeoffOption(BaseModel):
    option_name : str
    expected_cost : float
    expected_duration : int # In seconds

class MetricInput(BaseModel):
    job_id: str
    metric_name: str
    metric_value: float




from lotgi._openai import OpenAI

__all__ = [
    "OpenAI",
]


from lotgi.rest_client import RestClient
from lotgi.tokens import estimate_token_usage
from lotgi.logging import init_logging
from lotgi.models import rest

import click
from tabulate import tabulate
from datetime import timedelta
from typing import Optional
import sys
import requests
import datetime
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper
import os
import hashlib
import binascii


@click.group
def cli():
    """
    Lotgi CLI tool.
    """
    init_logging()


def format_duration(duration_s : int) -> str:
    total_hours = duration_s // (60 * 60)
    minutes = (duration_s // 60) % 60
    seconds = duration_s % 60
    if total_hours > 0:
        return f"{total_hours}h{minutes}m{seconds}s"
    return f"{minutes}m{seconds}s"

def format_execution_options(options) -> str:
    columns = {
        "Tradeoff": [
            option.option_name for option in options
        ],
        "Estimated Price": [
            f"${option.expected_cost:.2f}" for option in options
        ],
        "Estimated Duration (s)": [
            format_duration(option.expected_duration) for option in options
        ],
    }
    return tabulate(columns, headers="keys")

def get_tradeoff() -> str:
    while True:
        selection = input(f"Select a tradeoff ({repr(rest.EXECUTION_TRADEOFFS)}): ")
        if selection in rest.EXECUTION_TRADEOFFS:
            return selection
        else:
            print(f"Invalid tradeoff. Select from {rest.EXECUTION_TRADEOFFS}")

def upload_file(client : RestClient, filename : str, path : str):
    assert path.endswith(".jsonl"), f"The path must be to a jsonl file, not {path}"
    url = client.get_upload_url(filename)
    file_size = os.stat(path).st_size
    with open(path, "rb") as f:
        with tqdm(total=file_size, unit="B", unit_scale=True, unit_divisor=1024) as t:
            wrapped_file = CallbackIOWrapper(t.update, f, "read")
            requests.put(url, data=wrapped_file)

def sha256_file(f) -> str:
    """
    Return a string of the sha256 hash of a file. Do it in a streaming fashion to be memory efficient.
    """
    sha256 = hashlib.sha256()
    to_hash = f.read(2**20)
    while to_hash:
        sha256.update(to_hash)
        to_hash = f.read(2**20)
    return binascii.hexlify(sha256.digest()).decode()

@cli.command
@click.option("--model", required=True, help="The model to run inference with.")
@click.option("--input-file", required=False, help="The name of the file (uploaded with `lotgi upload`)to run inference on.")
@click.option("--input-path", required=False, help="The url or path to a file to run inference on.")
@click.option("--field", required=True, help="The field within a row containing the data.")
@click.option("--tradeoff", required=False, default=None, help=f"The tradeoff ({rest.EXECUTION_TRADEOFFS}).")
@click.option("--prompt", required=False, help="Prompt to use for inference. Leave empty if prompt is already included in the dataset.")
@click.option("--max-output-tokens", required=False, help="The maximum number of output tokens. [default=32]. Changing this value may affect the job cost.")
@click.option("--regex", required=False, help="Force the output to match a given regular expression.")
def submit(
        model : str,
        input_file : str,
        input_path : str, field : str,
        tradeoff : Optional[str],
        prompt: Optional[str],
        max_output_tokens : Optional[int],
        regex : Optional[str],
):
    """
    Submit a new batch inference job.

    Your job file is expected to be formatted as a jsonl file. Each line corresponds to a request and should be a json object. Use the `--field` argument to specify which field of the json object contains the text to complete.

    For example, if a file stored at `example.com/dataset.jsonl` contains the contents

    \b\n
    ```
    {"my_completion_text": "hello"}
    {"my_completion_text": "lorem"}
    ```

    You can run 
        `lotgi submit --model mistralai/Mistral-7B-Instruct-v0.2 
            --input-url example.com/dataset.jsonl 
            --field "my_completion_text" 
            --tradeoff Cheapest 
            --prompt "Summarize the following text"`. 
    
    This will run the mistral-7b model against the inputs "hello" and "lorem" with the provided prompt, optimizing for cost.
    """
    client = RestClient()

    # if tradeoff is None:
    #     print(f"Scanning job to estimate costs...")
    #     job_execution_options_input = estimate_token_usage(input_file=input_file, prompt="", model=model, field=field)
    #     options = client.job_exection_options(job_execution_options_input=job_execution_options_input)

    #     print(format_execution_options(options))
    #     tradeoff = get_tradeoff()

    if (not (input_file or input_path)) or (input_file and input_path):
        print("Must specify exactly one of --input-file or --input-path", file=sys.stderr)
        sys.exit(1)

    if input_path:
        with open(input_path, "rb") as f:
            input_file = "sha256:" + sha256_file(f)
        upload_file(client, input_file, input_path)

    result = client.submit_job(
        model=model,
        input_file=input_file,
        field=field,
        target_cost=-1,
        target_deadline=-1,
        execution_tradeoff=tradeoff,
        max_output_tokens=max_output_tokens,
        regex=regex,
    )
    print("Success!")
    print(result)


@cli.command
def list():
    """
    List your batch inferences jobs.
    """
    client = RestClient()
    jobs = sorted(client.list_jobs(), key=lambda job: job.submission_time, reverse=True)

    to_display = {
        "Job ID": [job.job_id for job in jobs],
        "Model": [job.model for job in jobs],
        "Status": [job.state for job in jobs],
        "Submission time": [datetime.datetime.fromtimestamp(job.submission_time).strftime('%Y-%m-%d %H:%M:%S') for job in jobs],
        "Tradeoff": [job.execution_tradeoff for job in jobs],
        "Field": [job.field for job in jobs],
        "Input": [job.input_file for job in jobs],
    }

    print(tabulate(to_display, headers="keys"))


@cli.command
@click.argument("job_id", required=True)
def get(job_id : str):
    """
    Get the results of a specific job.

    Job results are written to stdout. (Human readable text/errors are written to stderr).

    To store your job results, consider redirecting the job output to a file. `lotg get <job_id> > saved_job_results.jsonl`
    """
    client = RestClient()
    url = None
    try:
        url = client.get_job_results(job_id)
    except KeyError:
        print("Job not found!", file=sys.stderr, flush=True)
        sys.exit(1)
    assert url is not None

    result = requests.get(url)
    if not result.ok:
        print(f"Couldn't download result from {url}. {result.status_code} : {result.text}")
        sys.exit(1)

    print(result.text)

def upload_file(client : RestClient, filename : str, path : str):
    assert path.endswith(".jsonl"), f"The path must be to a jsonl file, not {path}"
    url = client.get_upload_url(filename)
    file_size = os.stat(path).st_size
    with open(path, "rb") as f:
        with tqdm(total=file_size, unit="B", unit_scale=True, unit_divisor=1024) as t:
            wrapped_file = CallbackIOWrapper(t.update, f, "read")
            requests.put(url, data=wrapped_file)


@cli.command
@click.argument("filename", required=True)
@click.argument("path", required=True)
def upload(filename : str, path : str):
    """
    Upload a dataset/file to be used for inference.

    Usage: lotgi upload <filename> <path>

    filename: The filename used to reference this file when running jobs.

    path: The path the file is currently stored at.
    """
    client = RestClient()
    upload_file(client, filename, path)

